import abc
import os
import json
import dbm

from .. import util
from . import client as gw2client, entity as gw2entity


class Storage (abc.ABC):
    @abc.abstractmethod
    def store_schema_version (self, version):
        pass

    @abc.abstractmethod
    def schema_version (self):
        pass

    @abc.abstractmethod
    def store_raw (self, path, result):
        pass

    @abc.abstractmethod
    def exists_raw (self, path, api_id):
        pass

    @abc.abstractmethod
    def raw (self, path, api_id):
        pass

    @abc.abstractmethod
    def clear_raw (self):
        pass

    @abc.abstractmethod
    def store (self, entity):
        pass

    @abc.abstractmethod
    def from_api_id (self, entity_type, api_id):
        pass

    @abc.abstractmethod
    def all_from_id (self, entity_type, id_):
        pass

    def from_id (self, entity_type, id_, filters=()):
        entities = self.all_from_id(entity_type, id_)
        filters = list(filters)
        filters.extend(entity_type.filters())
        for filter_ in filters:
            entities = filter_(entities)
            if len(entities) == 1:
                return entities[0]
            if len(entities) == 0:
                break
        raise KeyError((entity_type, id_))

    @abc.abstractmethod
    def clear (self):
        pass


class FileStorage (Storage):
    _SCHEMA_VERSION_KEY = 'meta:version'

    def __init__ (self, path=None):
        if path is None:
            default_cache_path = os.path.join(os.path.expanduser('~'), '.cache')
            cache_path = os.environ.get('XDG_CACHE_HOME', default_cache_path)
            self.path = os.path.join(cache_path, 'gw2buildutil')
        else:
            self.path = path

        os.makedirs(self.path, exist_ok=True)
        self._raw_db = dbm.open(os.path.join(self.path, 'api-raw.db'), 'c')
        try:
            self._db = dbm.open(os.path.join(self.path, 'api.db'), 'c')
        except Exception:
            self._raw_db.close()

    def close (self):
        try:
            self._raw_db.close()
        finally:
            self._db.close()

    def __enter__ (self):
        return self

    def __exit__ (self, *args):
        self.close()

    def store_schema_version (self, version):
        self._raw_db[self._SCHEMA_VERSION_KEY] = version

    def schema_version (self):
        if self._SCHEMA_VERSION_KEY in self._raw_db:
            return self._raw_db[self._SCHEMA_VERSION_KEY].decode()
        else:
            return None

    def _api_id_key (self, path, api_id):
        return f'entity:{"/".join(path)}:{api_id}'

    def store_raw (self, path, result):
        self._raw_db[self._api_id_key(path, result['id'])] = json.dumps(result)

    def exists_raw (self, path, api_id):
        return self._api_id_key(path, api_id) in self._raw_db

    def raw (self, path, api_id):
        return json.loads(self._raw_db[self._api_id_key(path, api_id)])

    def clear_raw (self):
        for key in self._raw_db.keys():
            del self._raw_db[key]

    def _id_key (self, entity_type, id_):
        return (f'{entity_type.type_id()}:'
                f'id:{util.Identified.normalise_id(id_)}')

    def store (self, entity):
        entity_type = type(entity)

        for id_ in entity.ids:
            id_key = self._id_key(entity_type, id_)
            if id_key in self._db:
                data = json.loads(self._db[id_key])
            else:
                data = []
            data.append(entity.api_id)
            self._db[id_key] = json.dumps(tuple(set(data)))

    def from_api_id (self, entity_type, api_id):
        return entity_type.from_api(self.raw(entity_type.path(), api_id), self)

    def all_from_id (self, entity_type, id_):
        key = self._id_key(entity_type, id_)
        api_ids = json.loads(self._db[key])
        return [self.from_api_id(entity_type, api_id) for api_id in api_ids]

    def clear (self):
        for key in self._db.keys():
            del self._db[key]



class Identified:
    def __init__ (self, ids):
        ids = (ids,) if isinstance(ids, str) else ids
        self.id_ = Identified.normalise_id(ids[0])
        self.ids = {Identified.normalise_id(id_) for id_ in ids}

    @staticmethod
    def normalise_id (id_):
        return str(id_).lower()



import enum

from . import util


def _enum_id_lookup (enum_cls):
    lookup = {}
    for item in enum_cls:
        for id_ in item.value.ids:
            lookup[id_] = item
    return lookup


class GameMode (util.Identified):
    def __init__ (self, name, suitable_game_modes):
        util.Identified.__init__(self, name)
        self.name = name
        self.suitable_game_modes = tuple(suitable_game_modes)

    def __str__ (self):
        return self.name


class GameModes (enum.Enum):
    OPEN_WORLD = GameMode('Open World', ('open world', 'dungeons'))
    DUNGEONS = GameMode('Dungeons', ('dungeons',))
    FRACTALS = GameMode('Fractals',
                         ('unorganised fractals', 'dungeons', 'open world'))
    RAIDS = GameMode('Raids', ('casual raids', 'casual organised fractals'))
    PVP = GameMode('PvP', ('PvP',))
    WVW = GameMode('WvW', ('WvW',))

    @staticmethod
    def from_id (id_):
        return _game_modes_id_lookup[util.Identified.normalise_id(id_)]

_game_modes_id_lookup = _enum_id_lookup(GameModes)


class BuildMetadata:
    def __init__ (self, game_mode, profession, elite_spec, labels=()):
        self.game_mode = game_mode
        self.profession = profession
        self.elite_spec = elite_spec
        self.labels = tuple(labels)

    @property
    def title (self):
        spec = self.profession if self.elite_spec is None else self.elite_spec
        return (f'{self.game_mode.value.name} {spec.name}: '
                f'{", ".join(self.labels)}')


class TextBody:
    pass


class MarkdownBody (TextBody):
    def __init__ (self, text):
        self.text = text


class WeaponType (util.Identified):
    def __init__ (self, ids, hands):
        util.Identified.__init__(self, ids)
        self.hands = hands


class WeaponTypes (enum.Enum):
    GREATSWORD = WeaponType('greatsword', 2)
    HAMMER = WeaponType('hammer', 2)
    LONGBOW = WeaponType('longbow', 2)
    RIFLE = WeaponType('rifle', 2)
    SHORTBOW = WeaponType('shortbow', 2)
    STAFF = WeaponType('staff', 2)

    AXE = WeaponType('axe', 1)
    DAGGER = WeaponType('dagger', 1)
    MACE = WeaponType('mace', 1)
    PISTOL = WeaponType('pistol', 1)
    SWORD = WeaponType('sword', 1)
    SCEPTRE = WeaponType(('sceptre', 'scepter'), 1)
    FOCUS = WeaponType('focus', 1)
    SHIELD = WeaponType('shield', 1)
    TORCH = WeaponType('torch', 1)
    WARHORN = WeaponType('warhorn', 1)

    HARPOON = WeaponType(('harpoon', 'speargun'), 2)
    SPEAR = WeaponType('spear', 2)
    TRIDENT = WeaponType('trident', 2)

    @staticmethod
    def from_id (id_):
        return _weapon_types_id_lookup[util.Identified.normalise_id(id_)]

_weapon_types_id_lookup = _enum_id_lookup(WeaponTypes)


class WeaponHand (util.Identified):
    def __init__ (self, ids, hands):
        util.Identified.__init__(self, ids)
        self.hands = hands


class WeaponHands (enum.Enum):
    BOTH = WeaponHand('both', 2)
    MAIN = WeaponHand('main', 1)
    OFF = WeaponHand('off', 1)

    @staticmethod
    def from_id (id_):
        return _weapon_hands_id_lookup[util.Identified.normalise_id(id_)]

_weapon_hands_id_lookup = _enum_id_lookup(WeaponHands)


class UpgradeTiers (enum.Enum):
    MINOR = 'minor'
    MAJOR = 'major'
    SUPERIOR = 'superior'


class Sigil:
    def __init__ (self, name, tier):
        self.name = name
        self.tier = tier


class PvpSigil (util.Identified):
    def __init__ (self, name):
        util.Identified.__init__(self, name)
        self.name = name


class PvpSigils (enum.Enum):
    ABSORPTION = PvpSigil('Absorption')
    AGONY = PvpSigil('Agony')
    BATTLE = PvpSigil('Battle')
    CLEANSING = PvpSigil('Cleansing')
    COMPOUNDING = PvpSigil('Compounding')
    CONFUSION = PvpSigil('Confusion')
    COURAGE = PvpSigil('Courage')
    DOOM = PvpSigil('Doom')
    ENERGY = PvpSigil('Energy')
    ENHANCEMENT = PvpSigil('Enhancement')
    ESCAPE = PvpSigil('Escape')
    EXPLOITATION = PvpSigil('Exploitation')
    EXPOSURE = PvpSigil('Exposure')
    INTELLIGENCE = PvpSigil('Intelligence')
    MISERY = PvpSigil('Misery')
    OPPORTUNITY = PvpSigil('Opportunity')
    PERIL = PvpSigil('Peril')
    PURGING = PvpSigil('Purging')
    REVOCATION = PvpSigil('Revocation')
    RUTHLESSNESS = PvpSigil('Ruthlessness')
    SAVAGERY = PvpSigil('Savagery')
    SEPARATION = PvpSigil('Separation')
    SMOLDERING = PvpSigil('Smoldering')
    STAGNATION = PvpSigil('Stagnation')
    VENOM = PvpSigil('Venom')

    @staticmethod
    def from_id (id_):
        return _pvp_sigils_id_lookup[util.Identified.normalise_id(id_)]

_pvp_sigils_id_lookup = _enum_id_lookup(PvpSigils)


class Weapon:
    def __init__ (self, type_, hand, stats, sigils):
        self.type_ = type_
        self.hand = hand
        self.stats = stats
        self.sigils = tuple(sigils)

        if len(self.sigils) != self.hand.value.hands:
            raise ValueError('wrong number of sigils for '
                             '{}: {}'.format(self.hand, len(self.sigils)))


class Weapons:
    def __init__ (self, set1, set2=None):
        self.set1 = tuple(set1)
        self.set2 = None if set2 is None else tuple(set2)

        if sum(w.hand.value.hands for w in self.set1) != 2:
            raise ValueError('wrong number of weapons in set 1: '
                            '{}'.format(self.set1))
        if (self.set2 is not None and
            sum(w.hand.value.hands for w in self.set2) != 2
        ):
            raise ValueError('wrong number of weapons in set 2: '
                            '{}'.format(self.set1))


class Rune:
    def __init__ (self, name, tier):
        self.name = name
        self.tier = tier


class PvpRune (util.Identified):
    def __init__ (self, name):
        util.Identified.__init__(self, name)
        self.name = name


class PvpRunes (enum.Enum):
    ADVENTURE = PvpRune('Adventure')
    AIR = PvpRune('Air')
    ALTRUISM = PvpRune('Altruism')
    BALTHAZAR = PvpRune('Balthazar')
    DIVINITY = PvpRune('Divinity')
    DWAYNA = PvpRune('Dwayna')
    EARTH = PvpRune('Earth')
    EVASION = PvpRune('Evasion')
    EXUBERANCE = PvpRune('Exuberance')
    FIRE = PvpRune('Fire')
    GRENTH = PvpRune('Grenth')
    HOELBRAK = PvpRune('Hoelbrak')
    ICE = PvpRune('Ice')
    INFILTRATION = PvpRune('Infiltration')
    LEADERSHIP = PvpRune('Leadership')
    LYSSA = PvpRune('Lyssa')
    MELANDRU = PvpRune('Melandru')
    ORR = PvpRune('Orr')
    RADIANCE = PvpRune('Radiance')
    RAGE = PvpRune('Rage')
    RATA_SUM = PvpRune('Rata Sum')
    RESISTANCE = PvpRune('Resistance')
    SANCTUARY = PvpRune('Sanctuary')
    SCAVENGING = PvpRune('Scavenging')
    SPEED = PvpRune('Speed')
    STRENGTH = PvpRune('Strength')
    THORNS = PvpRune('Thorns')
    VAMPIRISM = PvpRune('Vampirism')
    AFFLICTED = PvpRune('Afflicted')
    ARISTOCRACY = PvpRune('Aristocracy')
    BAELFIRE = PvpRune('Baelfire')
    BERSERKER = PvpRune('Berserker')
    CENTAUR = PvpRune('Centaur')
    CHRONOMANCER = PvpRune('Chronomancer')
    CITADEL = PvpRune('Citadel')
    DAREDEVIL = PvpRune('Daredevil')
    DEADEYE = PvpRune('Deadeye')
    DOLYAK = PvpRune('Dolyak')
    DRAGONHUNTER = PvpRune('Dragonhunter')
    DRUID = PvpRune('Druid')
    EAGLE = PvpRune('Eagle')
    ELEMENTALIST = PvpRune('Elementalist')
    ENGINEER = PvpRune('Engineer')
    FIGHTER = PvpRune('Fighter')
    FIREBRAND = PvpRune('Firebrand')
    FLAME_LEGION = PvpRune('Flame Legion')
    FLOCK = PvpRune('Flock')
    FORGE = PvpRune('Forge')
    GROVE = PvpRune('Grove')
    GUARDIAN = PvpRune('Guardian')
    HERALD = PvpRune('Herald')
    HOLOSMITH = PvpRune('Holosmith')
    KRAIT = PvpRune('Krait')
    LYNX = PvpRune('Lynx')
    MAD_KING = PvpRune('Mad King')
    MESMER = PvpRune('Mesmer')
    MIRAGE = PvpRune('Mirage')
    MONK = PvpRune('Monk')
    NECROMANCER = PvpRune('Necromancer')
    NIGHTMARE = PvpRune('Nightmare')
    RANGER = PvpRune('Ranger')
    REAPER = PvpRune('Reaper')
    RENEGADE = PvpRune('Renegade')
    REVENANT = PvpRune('Revenant')
    SCHOLAR = PvpRune('Scholar')
    SCRAPPER = PvpRune('Scrapper')
    SOLDIER = PvpRune('Soldier')
    SOULBEAST = PvpRune('Soulbeast')
    SPELLBREAKER = PvpRune('Spellbreaker')
    SUNLESS = PvpRune('Sunless')
    SVANIR = PvpRune('Svanir')
    TEMPEST = PvpRune('Tempest')
    THIEF = PvpRune('Thief')
    TRAPPER = PvpRune('Trapper')
    TRAVELER = PvpRune('Traveler')
    UNDEAD = PvpRune('Undead')
    WARRIOR = PvpRune('Warrior')
    WATER = PvpRune('Water')
    WEAVER = PvpRune('Weaver')
    WURM = PvpRune('Wurm')

    @staticmethod
    def from_id (id_):
        return _pvp_runes_id_lookup[util.Identified.normalise_id(id_)]

_pvp_runes_id_lookup = _enum_id_lookup(PvpRunes)


class ArmourType (util.Identified):
    def __init__ (self, ids):
        util.Identified.__init__(self, ids)


class ArmourTypes (enum.Enum):
    HELM = ArmourType(('helm', 'head'))
    SHOULDERS = ArmourType('shoulders')
    COAT = ArmourType(('coat', 'chest'))
    GLOVES = ArmourType(('gloves', 'hands'))
    LEGGINGS = ArmourType(('leggings', 'legs'))
    BOOTS = ArmourType(('boots', 'feet'))

    @staticmethod
    def from_id (id_):
        return _armour_types_id_lookup[util.Identified.normalise_id(id_)]

_armour_types_id_lookup = _enum_id_lookup(ArmourTypes)


class ArmourPiece:
    def __init__ (self, type_, stats, rune):
        self.type_ = type_
        self.stats = stats
        self.rune = rune


class Armour:
    def __init__ (self, pieces):
        self.pieces = {p.type_: p for p in pieces}

        if len(pieces) != 6:
            raise ValueError('expected 6 armour pieces, got '
                             '{}'.format(len(pieces)))
        if len(self.pieces) != 6:
            raise ValueError('not all armour types are present: '
                             '{}'.format(list(self.pieces.keys())))


class PvpArmour:
    def __init__ (self, rune):
        self.rune = rune


class TrinketTypes (enum.Enum):
    BACK = 'back'
    ACCESSORY_1 = 'accessory 1'
    ACCESSORY_2 = 'accessory 2'
    AMULET = 'amulet'
    RING_1 = 'ring 1'
    RING_2 = 'ring 2'


class Trinket:
    def __init__ (self, type_, stats):
        self.type_ = type_
        self.stats = stats


class Trinkets:
    def __init__ (self, pieces):
        self.pieces = {p.type_: p for p in pieces}

        if len(pieces) != 6:
            raise ValueError('expected 6 trinkets, got '
                             '{}'.format(len(pieces)))
        if len(self.pieces) != 6:
            raise ValueError('not all trinket types are present: '
                             '{}'.format(list(self.pieces.keys())))


class GearGroup (util.Identified):
    def __init__ (self, ids):
        util.Identified.__init__(self, ids)


class GearGroups (enum.Enum):
    WEAPONS = GearGroup('weapons')
    ARMOUR = GearGroup(('armour', 'armor'))
    TRINKETS = GearGroup('trinkets')
    ACCESSORIES = GearGroup('accessories')
    RINGS = GearGroup('rings')

    @staticmethod
    def from_id (id_):
        return _gear_groups_id_lookup[util.Identified.normalise_id(id_)]

_gear_groups_id_lookup = _enum_id_lookup(GearGroups)
_gear_groups_id_lookup.update(_weapon_types_id_lookup)
_gear_groups_id_lookup.update(_armour_types_id_lookup)
_gear_groups_id_lookup.update({type_.value: type_ for type_ in TrinketTypes})


class Consumables:
    def __init__ (self, food=None, utility=None):
        self.food = food
        self.utility = utility


class Gear:
    def __init__ (self, weapons, armour, trinkets, consumables):
        self.weapons = weapons
        self.armour = armour
        self.trinkets = trinkets
        self.consumables = consumables


class PvpGear:
    def __init__ (self, stats, weapons, armour):
        self.stats = stats
        self.weapons = weapons
        self.armour = armour


class TraitChoice:
    def __init__ (self, index, name):
        self.index = index
        self.name = name


class TraitChoices (enum.Enum):
    TOP = TraitChoice(0, 'Top')
    MIDDLE = TraitChoice(1, 'Middle')
    BOTTOM = TraitChoice(2, 'Bottom')

    @staticmethod
    def from_index (index):
        return _trait_choices_index_lookup[index]

_trait_choices_index_lookup = {choice.value.index: choice
                               for choice in TraitChoices}


class SpecialisationChoices:
    def __init__ (self, spec, choices):
        self.spec = spec
        self.choices = tuple(choices) # items can be None

        if len(self.choices) != 3:
            raise ValueError('expected 3 trait choices, got '
                             f'{len(self.choices)}')


class Traits:
    def __init__ (self, specs):
        self.specs = tuple(specs) # items can be None

        if len(self.specs) != 3:
            raise ValueError('expected 3 specialisations, got '
                             f'{len(self.specs)}')


class Skills:
    def __init__ (self, heal, utilities, elite):
        # skills can be None
        self.heal = heal
        self.utilities = tuple(utilities)
        self.elite = elite

        if len(self.utilities) != 3:
            raise ValueError('expected 3 utility skills, got '
                             f'{len(self.utilities)}')


class RevenantSkills (Skills):
    def __init__ (self, legends):
        self.legends = tuple(legends) # items can be None

        if len(self.legends) != 2:
            raise ValueError(f'expected 2 legends, got {len(self.legends)}')


class RangerPets:
    def __init__ (self, pets):
        self.pets = tuple(pets) # items can be None

        if len(self.pets) < 1 or len(self.pets) > 2:
            raise ValueError(f'expected 1 or 2 pets, got {len(self.pets)}')


class RangerOptions:
    def __init__ (self, pets, aquatic_pets=None):
        self.pets = pets
        self.aquatic_pets = (RangerPets((None, None))
                             if aquatic_pets is None else aquatic_pets)


class Intro:
    def __init__ (self, url, description, gear, traits, skills,
                  profession_options=None, aquatic_skills=None):
        self.url = url
        self.description = description
        self.gear = gear
        self.traits = traits
        self.skills = skills
        self.profession_options = profession_options
        self.aquatic_skills = aquatic_skills


class Boon (util.Identified):
    def __init__ (self, name, extra_ids=()):
        ids = (name,) + (
            (extra_ids,) if isinstance(extra_ids, str) else tuple(extra_ids))
        util.Identified.__init__(self, ids)
        self.name = name


class Boons (enum.Enum):
    AEGIS = Boon('Aegis')
    ALACRITY = Boon('Alacrity')
    FURY = Boon('Fury')
    MIGHT = Boon('Might')
    PROTECTION = Boon('Protection')
    QUICKNESS = Boon('Quickness')
    REGENERATION = Boon('Regeneration')
    RESISTANCE = Boon('Resistance')
    RETALIATION = Boon('Retaliation')
    STABILITY = Boon('Stability')
    SWIFTNESS = Boon('Swiftness')
    VIGOUR = Boon('Vigour', ('vigor',))

    @staticmethod
    def from_id (id_):
        return _boons_id_lookup[util.Identified.normalise_id(id_)]

_boons_id_lookup = _enum_id_lookup(Boons)


class BoonTarget (util.Identified):
    def __init__ (self, num, name):
        util.Identified.__init__(self, (str(num), name))
        self.num = num
        self.name = name


class BoonTargets (enum.Enum):
    PARTY = BoonTarget(5, 'party')
    SQUAD = BoonTarget(10, 'squad')

    @staticmethod
    def from_id (id_):
        return _boon_targets_id_lookup[util.Identified.normalise_id(id_)]

_boon_targets_id_lookup = _enum_id_lookup(BoonTargets)


class BoonUptime:
    def __init__ (self, boon, target, uptime_percent):
        self.boon = boon
        self.target = target
        self.uptime_percent = uptime_percent


class BoonUptimeVariant:
    def __init__ (self, boon_uptimes):
        self.boon_uptimes = boon_uptimes


class BoonNotes:
    def __init__ (self, boon_uptime_variants):
        self.boon_uptime_variants = boon_uptime_variants


class Build:
    def __init__ (self, metadata, intro, alternatives=None, usage=None,
                  notes=None, boon_notes=None, encounters=None):
        self.metadata = metadata
        self.intro = intro
        self.alternatives = alternatives
        self.usage = usage
        self.notes = notes
        self.boon_notes = boon_notes
        self.encounters = encounters



import json
import urllib.parse
import urllib.request

BASE_URL = 'https://api.guildwars2.com/v2'
SCHEMA_VERSION = '2020-10-01T00:00:00Z'
MAX_QUERYSTRING_SIZE = 1024
BATCH_SIZE = 100


class Client:
    def __init__ (self, base_url=BASE_URL, batch_size=BATCH_SIZE):
        self.base_url = base_url
        self.batch_size = batch_size

    def _url (self, path):
        path = '/'.join(urllib.parse.quote(part, safe='') for part in path)
        return f'{self.base_url}/{path}'

    def _get_json (self, url):
        req = urllib.request.Request(url, headers={
            'Accept': 'application/json',
            'X-Schema-Version': SCHEMA_VERSION,
        })
        with urllib.request.urlopen(req) as res:
            body = res.read()
        return json.loads(body)

    def list_(self, path):
        return self._get_json(self._url(path))

    def _get_batch (self, path, querystring_parts):
        url = f'{self._url(path)}?{"".join(querystring_parts)}'
        return self._get_json(url)

    def get (self, path, ids):
        sep = urllib.parse.quote_plus(',')
        sep_size = len(sep)

        querystring_parts = ['ids=']
        batch_size = 0
        querystring_size = 4
        for id_ in ids:
            id_quoted = urllib.parse.quote_plus(str(id_))
            if batch_size > 0 and (
                batch_size + 1 > BATCH_SIZE or
                querystring_size + len(id_quoted) > MAX_QUERYSTRING_SIZE
            ):
                yield from self._get_batch(path, querystring_parts)
                querystring_parts = ['ids=']
                batch_size = 0
                querystring_size = 4
            querystring_parts.append(sep)
            querystring_parts.append(id_quoted)
            batch_size += 1
            querystring_size += sep_size + len(id_quoted)

        if batch_size > 0:
            yield from self._get_batch(path, querystring_parts)



import io
import re

from .. import api, build
from . import util, text as parse_text, section

_SKIP_SECTION = object()


title_pattern = re.compile('^'
    f'(?P<mode>{parse_text.words_pattern}) '
    f'(?P<prof>{parse_text.word_pattern}) '
    '\\((?P<labels>'
         f'{parse_text.sep_pattern(", ", parse_text.words_pattern)}'
         ')\\)'
    '$')

def parse_title (title, api_storage):
    match = title_pattern.match(title)
    if match is None:
        raise util.ParseError('title doesn\'t match expected format: '
                              '{}'.format(repr(title)))
    fields = match.groupdict()

    game_mode = build.GameModes.from_id(fields['mode'])
    if game_mode is None:
        raise util.ParseError('title doesn\'t start with a known '
                              'game modes identifier: {}'.format(repr(title)))

    try:
        profession = api_storage.from_id(api.entity.Profession, fields['prof'])
    except KeyError:
        try:
            elite_spec = api_storage.from_id(
                api.entity.Specialisation, fields['prof'])
        except KeyError:
            elite_spec = None
        if elite_spec is None or not elite_spec.is_elite:
            raise util.ParseError('title has a missing or incorrect profession '
                                  f'identifier: {repr(title)}')
        else:
            profession = elite_spec.profession
    else:
        elite_spec = None

    labels = [l.strip() for l in fields['labels'].split(',')]
    return build.BuildMetadata(game_mode, profession, elite_spec, labels)


section_parsers = {
    'intro': section.intro,
    'alternatives': section.alternatives,
    'usage': section.usage,
    'notes': section.notes,
    'boon notes': section.boonnotes,
    'encounters': section.encounters,
    'calculation': _SKIP_SECTION,
}


def parse_body (f, meta, api_storage):
    lines = (line if isinstance(line, str) else line.decode('utf-8')
             for line in f)
    build_data = {}

    for title, section_lines in util.split_sections(lines, 'intro'):
        section_module = section_parsers.get(title)
        if section_module is None:
            raise util.ParseError('unknown section: {}'.format(repr(title)))
        elif section_module is _SKIP_SECTION:
            for line in section_lines:
                pass
        else:
            build_data[title.replace(' ', '_')] = (
                section_module.parse(section_lines, meta, api_storage))

    return build.Build(meta, **build_data)



import re

from ... import build
from .. import util


uptimes_pattern = re.compile('^'
    '(\d{1,3}% )?(5|10)-[a-z]+(, (\d{1,3}% )?(5|10)-[a-z]+)*'
    '$')

def parse_uptime (line):
    uptimes = []
    if uptimes_pattern.match(line) is None:
        raise util.ParseError('boon uptime definition doesn\'t match expected '
                              'format: {}'.format(repr(line)))

    for boon_uptime_text in line.split(', '):
        if ' ' in boon_uptime_text:
            uptime_text, rest = boon_uptime_text.split(' ')
        else:
            uptime_text = None
            rest = boon_uptime_text
        target_text, boon_text = rest.split('-')

        uptime = None if uptime_text is None else int(uptime_text[:-1])
        target = build.BoonTargets.from_id(target_text)
        if target is None:
            raise util.ParseError(
                'boon uptime definition has a missing or incorrect target '
                'count identifier: {}'.format(repr(boon_uptime_text)))
        boon = build.Boons.from_id(boon_text)
        if boon is None:
            raise util.ParseError(
                'boon uptime definition has a missing or incorrect boon '
                'identifier: {}'.format(repr(boon_uptime_text)))
        uptimes.append(build.BoonUptime(boon, target, uptime))

    uptimes.sort(key = lambda uptime: (uptime.boon.name, uptime.target.name))
    return uptimes


def parse (lines, meta, api_storage):
    paragraphs = list(util.group_paragraphs(
        util.strip_empty_lines(lines, inner='collapse')))
    if len(paragraphs) < 1:
        raise util.ParseError('boon notes is incomplete')

    boon_variants = [build.BoonUptimeVariant(parse_uptime(line))
                     for line in paragraphs[0]]

    # TODO: other lines
    return build.BoonNotes(boon_variants)



import abc
import inspect
import re

from .. import build, util


class Entity (abc.ABC, util.Identified):
    def __init__ (self, api_id, ids):
        util.Identified.__init__ (self, ids)
        self.api_id = api_id

    def __eq__ (self, other):
        return type(self) is type(other) and self.api_id == other.api_id

    @classmethod
    def type_id (cls):
        # must not contain ':'
        return cls.__name__.lower()

    @staticmethod
    def crawl_dependencies ():
        # just a performance hint - should crawl dependencies in from_api
        return set()

    @staticmethod
    @abc.abstractmethod
    def path ():
        pass

    @staticmethod
    @abc.abstractmethod
    def from_api (result, storage, crawler=None):
        # return instance, or None to skip
        pass

    @staticmethod
    def _filter_use_first (entities):
        # filter to entity with earliest API ID (for determinism), for each
        # unique name
        by_name = {}
        for e in entities:
            by_name.setdefault(e.name, []).append(e)
        return [sorted(group, key=lambda e: e.api_id)[0]
                for group in by_name.values()]

    @staticmethod
    def filters ():
        return [Entity._filter_use_first]


class Profession (Entity):
    def __init__ (self, api_id, name, build_id, skills_build_ids):
        Entity.__init__(self, api_id, (name, build_id))
        self.name = name
        self.build_id = build_id
        self.skills_build_ids = skills_build_ids

    @staticmethod
    def path ():
        return ('professions',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        skills_build_ids = {
            api_id: build_id
            for build_id, api_id in result.get('skills_by_palette', [])}
        return Profession(
            result['id'], result['name'], result['code'], skills_build_ids)


class Specialisation (Entity):
    def __init__ (self, api_id, name, profession, is_elite):
        Entity.__init__(self, api_id, name)
        self.name = name
        self.profession = profession
        self.is_elite = is_elite

    @staticmethod
    def crawl_dependencies ():
        return set([Profession])

    @staticmethod
    def path ():
        return ('specializations',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        prof_api_id = result['profession']
        if crawler is not None:
            crawler.crawl(Profession, (prof_api_id,))
        prof = storage.from_api_id(Profession, prof_api_id)
        return Specialisation(
            result['id'], result['name'], prof, result['elite'])


_skill_name_mappings = {
    'Mirage Mirror': 'Jaunt',
}

class Skill (Entity):
    def __init__ (self, api_id, name, build_id, storage_build_id, is_chained):
        full_id = name.lower().strip('"')
        ids = []
        id_ = full_id
        if id_.endswith('!'):
            id_ = id_[:-1]
        if ':' in id_:
            id_ = id_[id_.index(':') + 1:].strip()
        if '.' in id_:
            id_ = id_.replace('.', '') # eg. 'A.E.D.'
        if id_.startswith('summon '):
            id_ = id_[len('summon '):]

        ids = [id_]
        if full_id != id_:
            ids.append(full_id)
        if storage_build_id is not None:
            ids.append(storage_build_id)
        if ' ' in id_:
            abbr = ''.join(word[0] for word in id_.split(' ') if word)
            ids.append(abbr)
            if full_id.endswith('!'):
                ids.append(abbr + '!')

        Entity.__init__(self, api_id, ids)
        self.name = name
        self.build_id = build_id
        self.is_chained = is_chained

    @staticmethod
    def crawl_dependencies ():
        return set([Profession])

    @staticmethod
    def path ():
        return ('skills',)

    @staticmethod
    def _storage_build_id (profession, build_id):
        return f'build:{profession.api_id}:{build_id}'

    @staticmethod
    def from_api (result, storage, crawler=None):
        api_id = result['id']
        name = result['name']
        name = _skill_name_mappings.get(name, name)
        build_id = None
        storage_build_id = None

        prof_api_ids = result.get('professions', ())
        if len(prof_api_ids) == 1:
            if crawler is not None:
                crawler.crawl(Profession, (prof_api_ids[0],))
            prof = storage.from_api_id(Profession, prof_api_ids[0])
            build_id = prof.skills_build_ids.get(api_id)
            if build_id is not None:
                storage_build_id = Skill._storage_build_id(prof, build_id)

        return Skill(api_id, name, build_id, storage_build_id,
                     is_chained='prev_chain' in result)

    @staticmethod
    def from_build_id (profession, build_id, storage):
        storage_build_id = Skill._storage_build_id(profession, build_id)
        return storage.from_id(Skill, storage_build_id)

    @staticmethod
    def _filter_not_chained (skills):
        return [s for s in skills if not s.is_chained]

    @staticmethod
    def _filter_has_build_id (skills):
        return [s for s in skills if s.build_id is not None]

    @staticmethod
    def filters ():
        return [
            Skill._filter_not_chained,
            Skill._filter_has_build_id,
        ] + Entity.filters()


# not obtainable through the API in any sensible way
_legend_ids = {
    'assassin': ('shiro',),
    'centaur': ('ventari',),
    'demon': ('mallyx',),
    'dwarf': ('jalis',),

    'dragon': ('glint', 'herald'),
    'renegade': ('kalla',),
}

class RevenantLegend (Entity):
    def __init__ (self, api_id, name, build_id, skills):
        id_ = name.lower()
        ids = [id_]
        ids.extend(_legend_ids.get(id_, ()))
        ids.append(build_id)

        Entity.__init__(self, api_id, ids)
        self.name = name
        self.build_id = build_id
        self.heal_skill = skills['heal']
        self.utility_skills = tuple(skills['utilities'])
        self.elite_skill = skills['elite']

    @staticmethod
    def crawl_dependencies ():
        return set([Skill])

    @staticmethod
    def path ():
        return ('legends',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if crawler is not None:
            crawler.crawl(Skill, (result['swap'],))
        swap_skill = storage.from_api_id(Skill, result['swap'])
        name = swap_skill.name
        if name.startswith('Legendary '):
            name = name[len('Legendary '):]
        if name.endswith(' Stance'):
            name = name[:-len(' Stance')]

        def lookup_skill (api_id):
            if crawler is not None:
                crawler.crawl(Skill, (api_id,))
            return storage.from_api_id(Skill, api_id)

        skills = {
            'heal': lookup_skill(result['heal']),
            'utilities': [lookup_skill(api_id)
                          for api_id in result['utilities']],
            'elite': lookup_skill(result['elite']),
        }

        return RevenantLegend(result['id'], name, result['code'], skills)


class RangerPet (Entity):
    def __init__ (self, api_id, name):
        full_id = name.lower()
        id_ = full_id
        if id_.startswith('juvenile '):
            id_ = id_[len('juvenile '):]
        ids = [id_]
        if full_id != id_:
            ids.append(full_id)

        Entity.__init__(self, api_id, ids)
        self.name = name

    @staticmethod
    def path ():
        return ('pets',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        return RangerPet(result['id'], result['name'])


class Stats (Entity):
    def __init__ (self, api_id, name, num_attributes):
        full_id = name
        id_ = full_id.replace('\'s', '')
        ids = [id_]
        if full_id != id_:
            ids.append(full_id)

        Entity.__init__(self, api_id, ids)
        self.name = name
        self.num_attributes = num_attributes

    @staticmethod
    def path ():
        return ('itemstats',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if not result['name']:
            return None

        return Stats(result['id'], result['name'],
                     num_attributes=len(result['attributes']))

    @staticmethod
    def _filter_endgame (stat_sets):
        return [stats for stats in stat_sets if stats.num_attributes >= 3]

    @staticmethod
    def _filter_not_mixed (stat_sets):
        return [stats for stats in stat_sets if stats.name.find(' and ') < 0]

    @staticmethod
    def filters ():
        return [
            Stats._filter_endgame,
            Stats._filter_not_mixed,
        ] + Entity.filters()


class PvpStats (Entity):
    def __init__ (self, api_id, name):
        id_ = name.lower()
        if id_.endswith(' amulet'):
            id_ = id_[:-len(' amulet')]

        Entity.__init__(self, api_id, id_)
        self.name = name

    @staticmethod
    def path ():
        return ('pvp', 'amulets')

    @staticmethod
    def from_api (result, storage, crawler=None):
        return PvpStats(result['id'], result['name'])


sigil_pattern = re.compile(r'^'
    r'(?P<tier>\w+) Sigil of (the )?(?P<name>[\w\' ]+)'
    r'$')

class Sigil (Entity):
    def __init__ (self, api_id, name):
        match = sigil_pattern.match(name)
        if match is None:
            raise ValueError(f'unexpected sigil name format: {name}')
        fields = match.groupdict()
        tier = build.UpgradeTiers(fields['tier'].lower())
        ids = [f'{tier.value} {fields["name"]}']
        if tier == build.UpgradeTiers.SUPERIOR:
            ids.append(fields['name'])

        Entity.__init__(self, api_id, ids)
        self.name = name
        self.tier = tier

    @staticmethod
    def path ():
        return ('items',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if result['type'] != 'UpgradeComponent':
            return None
        if result['details']['type'] != 'Sigil':
            return None
        if result['name'] == 'Legendary Sigil':
            return None

        return Sigil(result['id'], result['name'])


rune_pattern = re.compile(r'^'
    r'(?P<tier>\w+) Rune of (the )?(?P<name>[\w\' ]+)'
    r'$')

class Rune (Entity):
    def __init__ (self, api_id, name):
        match = rune_pattern.match(name)
        if match is None:
            raise ValueError(f'unexpected rune name format: {name}')
        fields = match.groupdict()
        tier = build.UpgradeTiers(fields['tier'].lower())
        ids = [f'{tier.value} {fields["name"]}']
        if tier == build.UpgradeTiers.SUPERIOR:
            ids.append(fields['name'])

        Entity.__init__(self, api_id, ids)
        self.name = name
        self.tier = tier

    @staticmethod
    def path ():
        return ('items',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if result['type'] != 'UpgradeComponent':
            return None
        if result['details']['type'] != 'Rune':
            return None
        if result['name'] in ('', 'Legendary Rune'):
            return None

        return Rune(result['id'], result['name'])


food_prefixes = (
    'plate', 'bowl', 'can', 'pot', 'cup', 'jug', 'bit', 'slice', 'loaf',
    'strip', 'glass', 'handful', 'piece',
)

class Food (Entity):
    def __init__ (self, api_id, name):
        full_id = name.lower()
        ids = [full_id]
        for prefix in food_prefixes:
            full_prefix = f'{prefix} of '
            if full_id.startswith(full_prefix):
                ids.append(full_id[len(full_prefix):])
                break

        Entity.__init__(self, api_id, ids)
        self.name = name

    @staticmethod
    def path ():
        return ('items',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if result['type'] != 'Consumable':
            return None
        if result['details']['type'] != 'Food':
            return None

        return Food(result['id'], result['name'])


class UtilityConsumable (Entity):
    def __init__ (self, api_id, name):
        Entity.__init__(self, api_id, name)
        self.name = name

    @staticmethod
    def path ():
        return ('items',)

    @staticmethod
    def from_api (result, storage, crawler=None):
        if result['type'] != 'Consumable':
            return None
        if result['details']['type'] != 'Utility':
            return None

        return UtilityConsumable(result['id'], result['name'])


BUILTIN_TYPES = [
    o for o in vars().values()
    if inspect.isclass(o) and issubclass(o, Entity) and o is not Entity]



import logging

from . import client as gw2client, entity as gw2entity, storage as gw2storage

logger = logging.getLogger(__name__)


def _dependency_order (entity_types):
    remaining = set(entity_types)
    while remaining:
        used = set()
        for t in remaining:
            if not t.crawl_dependencies() & remaining:
                yield t
                used.add(t)
        if not used:
            yield from remaining
            break
        remaining -= used


class Crawler:
    def __init__ (self, client, storage, entity_types):
        self.client = client
        self.storage = storage
        self.entity_types = entity_types

    def _crawl (self, path, api_ids):
        new_api_ids = [api_id for api_id in api_ids
                       if not self.storage.exists_raw(path, api_id)]
        if not new_api_ids:
            return
        logger.info(f'get {len(new_api_ids)}/{len(api_ids)} /{"/".join(path)}')

        for result in self.client.get(path, new_api_ids):
            self.storage.store_raw(path, result)

    def _process (self, path, api_ids, entity_types):
        for api_id in api_ids:
            for entity_type in entity_types:
                result = self.storage.raw(path, api_id)
                entity = entity_type.from_api(result, self.storage, self)
                if entity is not None:
                    self.storage.store(entity)

    def crawl (self, entity_type, api_ids):
        path = entity_type.path()
        self._crawl(path, api_ids)
        self._process(path, api_ids, (entity_type,))

    def crawl_all (self, entity_types):
        by_path = {}
        for entity_type in entity_types:
            by_path.setdefault(entity_type.path(), []).append(entity_type)

        ordered = list(_dependency_order(entity_types))
        ordered_grouped = sorted(by_path.values(),
                                 key=lambda group: ordered.index(group[0]))

        for group in ordered_grouped:
            path = group[0].path()
            logger.info(f'list /{"/".join(path)}')
            api_ids = self.client.list_(path)
            self._crawl(path, api_ids)
            self._process(path, api_ids, group)


def crawl (client=gw2client.Client(),
           storage=None,
           entity_types=gw2entity.BUILTIN_TYPES,
           full_recrawl=False):
    if storage is None:
        with gw2storage.FileStorage() as storage:
            crawl(client, storage, entity_types)
        return

    crawler = Crawler(client, storage, entity_types)
    if storage.schema_version() != gw2client.SCHEMA_VERSION or full_recrawl:
        storage.clear_raw()
        storage.store_schema_version(gw2client.SCHEMA_VERSION)
    storage.clear() # required if entity definition changes
    crawler.crawl_all(entity_types)



import re
import collections

from ... import build
from ... import api
from .. import util, text as parse_text

wds_pat = parse_text.words_pattern


url_pattern = re.compile(r'^http://gw2skills.net/editor/\?[A-Za-z0-9+-/]+$')

def parse_url (lines):
    if len(lines) > 1:
        raise util.ParseError('first intro paragraph should have only 1 line')
    if url_pattern.match(lines[0]) is None:
        raise util.ParseError('first intro line doesn\'t look like a valid URL')
    return lines[0]


def parse_description (paragraphs):
    return build.MarkdownBody(
        '\n\n'.join('\n'.join(lines) for lines in paragraphs))


stats_pattern = re.compile('^'
    f'{parse_text.sep_pattern(", ", parse_text.sep_pattern(" + ", wds_pat))}'
    '$')

def parse_stats (line, meta, api_storage):
    if stats_pattern.match(line) is None:
        raise util.ParseError('stats definition doesn\'t match expected '
                              'format: {}'.format(repr(line)))

    stats_entity_type = (
        api.entity.PvpStats if meta.game_mode is build.GameModes.PVP
        else api.entity.Stats)
    gear_stats = {}
    for section in line.split(', '):
        words = section.split()
        try:
            stats = api_storage.from_id(stats_entity_type, words[0])
        except KeyError:
            raise util.ParseError(f'unknown stats: {words[0]}')
        gear_groups = parse_text.parse_gear_groups(section[len(words[0]) + 1:])

        if not gear_groups:
            gear_stats[None] = stats
        for gear_group in gear_groups:
            gear_stats[gear_group] = stats

    if None not in gear_stats:
        raise util.ParseError('no default stats found')
    return gear_stats


def stats_lookup (names, stats):
    for name in names:
        if name in stats:
            return stats[name]
    return stats[None]


def lookup_sigil (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.Sigil, id_)
    except KeyError:
        raise util.ParseError(f'unknown sigil: {id_}')


weapons_pattern = re.compile('^'
    f'(?P<types1>{wds_pat}) \\((?P<sigils1>{wds_pat}, {wds_pat})\\)'
    f'( / (?P<types2>{wds_pat}) \\((?P<sigils2>{wds_pat}, {wds_pat})\\))?'
    '$')

def parse_weapons (line, stats, meta, api_storage):
    match = weapons_pattern.match(line)
    if match is None:
        raise util.ParseError('weapons definition doesn\'t match expected '
                              'format: {}'.format(repr(line)))
    fields = match.groupdict()
    parse_sigil = (
        build.PvpSigils.from_id if meta.game_mode == build.GameModes.PVP
        else lambda text: lookup_sigil(text, api_storage))

    def build_weapon (type_, hand, sigils):
        weapon_stats = stats_lookup((type_, build.GearGroups.WEAPONS), stats)
        return build.Weapon(type_, hand, weapon_stats, sigils)

    def build_weapon_set (weapons_field, sigils_field):
        types = [build.WeaponTypes.from_id(t) for t in weapons_field.split()]
        sigils = [parse_sigil(s) for s in sigils_field.split(', ')]
        if len(types) == 2:
            return (
                build_weapon(types[0], build.WeaponHands.MAIN, (sigils[0],)),
                build_weapon(types[1], build.WeaponHands.OFF, (sigils[1],)))
        else:
            return (build_weapon(types[0], build.WeaponHands.BOTH, sigils),)

    return build.Weapons(
        build_weapon_set(fields['types1'], fields['sigils1']),
        (build_weapon_set(fields['types2'], fields['sigils2'])
         if fields['types2'] is not None else None))


def lookup_rune (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.Rune, id_)
    except KeyError:
        raise util.ParseError(f'unknown rune: {id_}')


runes_pattern = re.compile('^('
    f'(?P<single>{wds_pat})'
    '|'
    f'(?P<multi>\\d {wds_pat}( \\+ \\d {wds_pat})+)'
    ') runes$')

def parse_runes (runes_line):
    match = runes_pattern.match(runes_line)
    if match is None:
        raise util.ParseError('runes definition doesn\'t match expected '
                              'format: {}'.format(repr(runes_line)))

    fields = match.groupdict()
    if fields['single'] is not None:
        runes = collections.Counter({fields['single']: 6})
    else:
        rune_items = (item.split(' ', 1)
                      for item in fields['multi'].split(' + '))
        runes = collections.Counter(
            {type_: int(count) for count, type_ in rune_items})
    if sum(runes.values()) != 6:
        raise util.ParseError('wrong total rune count: {}'.format(dict(runes)))
    return runes


def parse_armour (runes, stats, api_storage):
    return build.Armour([build.ArmourPiece(
        type_,
        stats_lookup((type_, build.GearGroups.ARMOUR), stats),
        lookup_rune(rune, api_storage)
    ) for type_, rune in zip(build.ArmourTypes, runes.elements())])


def parse_pvp_armour (runes, stats):
    if len(runes) > 1:
        raise util.ParseError('different runes are not allowed in PvP')

    return build.PvpArmour(build.PvpRunes.from_id(next(iter(runes))))


def parse_trinkets (stats):
    T = build.TrinketTypes
    G = build.GearGroups
    return build.Trinkets([
        build.Trinket(T.BACK, stats_lookup(
            (T.BACK, G.TRINKETS), stats)),
        build.Trinket(T.ACCESSORY_1, stats_lookup(
            (T.ACCESSORY_1, G.ACCESSORIES, G.TRINKETS), stats)),
        build.Trinket(T.ACCESSORY_2, stats_lookup(
            (T.ACCESSORY_2, G.ACCESSORIES, G.TRINKETS), stats)),
        build.Trinket(T.AMULET, stats_lookup(
            (T.AMULET, G.TRINKETS), stats)),
        build.Trinket(T.RING_1, stats_lookup(
            (T.RING_1, G.RINGS, G.TRINKETS), stats)),
        build.Trinket(T.RING_2, stats_lookup(
            (T.RING_2, G.RINGS, G.TRINKETS), stats)),
    ])


traits_pattern = re.compile('^'
    f'{wds_pat}' '( [1-3]){3}'
    f'(, {wds_pat}' '( [1-3]){3}){2}'
    '$')

def parse_traits (line, api_storage):
    if traits_pattern.match(line) is None:
        raise util.ParseError('traits definition doesn\'t match expected '
                              f'format: {repr(line)}')

    specs = []
    for spec_text in line.split(', '):
        id_, *choices_text = spec_text.rsplit(' ', 3)
        choices = [build.TraitChoices.from_index(int(choice_text) - 1)
                   for choice_text in choices_text]
        try:
            spec = api_storage.from_id(api.entity.Specialisation, id_)
        except KeyError:
            raise util.ParseError(f'unknown specialisation: {id_}')
        specs.append(build.SpecialisationChoices(spec, choices))

    return build.Traits(specs)


def lookup_consumable (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.Food, id_)
    except KeyError:
        try:
            return api_storage.from_id(api.entity.UtilityConsumable, id_)
        except KeyError:
            raise util.ParseError(f'unknown consumable: {id_}')


def parse_consumables (line, api_storage):
    consumables_text = parse_text.parse_words_seq(line, 'consumables')
    if not consumables_text or len(consumables_text) > 2:
        raise util.ParseError('expected 1 or 2 consumables, got '
                              f'{len(consumables_text)}')

    consumables = [lookup_consumable(text, api_storage)
                   for text in consumables_text]
    by_type = {}
    for c in consumables:
        t = type(c)
        if t in by_type:
            raise util.ParseError(f'multiple {t.type_id()} consumables '
                                  f'specified: {by_type[t].name}, {c.name}')
        by_type[t] = c

    return build.Consumables(by_type.get(api.entity.Food),
                             by_type.get(api.entity.UtilityConsumable))


num_prof_lines = collections.defaultdict(lambda: 0, {
    'ranger': 1,
})


def lookup_ranger_pet (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.RangerPet, id_)
    except KeyError:
        raise util.ParseError(f'unknown ranger pet: {id_}')


def parse_ranger_options (line, api_storage):
    ids = parse_text.parse_words_seq(line, 'ranger pets')
    pets = build.RangerPets(
        [lookup_ranger_pet(id_, api_storage) for id_ in ids])
    return build.RangerOptions(pets)


def parse_prof_options (lines, meta, api_storage):
    if meta.profession.id_ == 'ranger':
        return parse_ranger_options(lines[0], api_storage)
    else:
        return None


def parse_setup (lines, meta, api_storage):
    prof_lines = num_prof_lines[meta.profession.id_]
    min_lines = 4 + prof_lines
    max_lines = 5 + prof_lines
    num_lines = len(lines)
    if num_lines not in list(range(min_lines, max_lines + 1)):
        raise util.ParseError(
            'second intro paragraph has the wrong number of lines: '
            f'got {num_lines}, expected {min_lines}-{max_lines}')
    if num_lines == max_lines and meta.game_mode == build.GameModes.PVP:
        raise util.ParseError('consumables are not allowed for PvP builds')

    stats = parse_stats(lines[1], meta, api_storage)
    if meta.game_mode == build.GameModes.PVP:
        gear = build.PvpGear(
            stats[None],
            parse_weapons(lines[0], stats, meta, api_storage),
            parse_pvp_armour(parse_runes(lines[2]), stats))
    else:
        gear = build.Gear(
            parse_weapons(lines[0], stats, meta, api_storage),
            parse_armour(parse_runes(lines[2]), stats, api_storage),
            parse_trinkets(stats),
            (parse_consumables(lines[4], api_storage)
            if num_lines == max_lines else build.Consumables()))

    return {
        'gear': gear,
        'traits': parse_traits(lines[3], api_storage),
        'profession options':
            parse_prof_options(lines[-prof_lines:], meta, api_storage),
    }


def lookup_revenant_legend (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.RevenantLegend, id_)
    except KeyError:
        raise util.ParseError(f'unknown Revenant legend: {id_}')


def parse_revenant_skills (lines, api_storage):
    if len(lines) != 1:
        raise util.ParseError(
            'third intro paragraph has the wrong number of lines: '
            f'got {len(lines)}, expected 1')

    legends_text = parse_text.parse_words_seq(lines[0], 'legends', 2)
    return build.RevenantSkills([
        lookup_revenant_legend(legend_text, api_storage)
        for legend_text in legends_text])


def lookup_skill (id_, api_storage):
    try:
        return api_storage.from_id(api.entity.Skill, id_)
    except KeyError:
        raise util.ParseError(f'unknown skill: {id_}')


def parse_skills (lines, meta, api_storage):
    if meta.profession.id_ == 'revenant':
        return parse_revenant_skills(lines, api_storage)

    if len(lines) != 3:
        raise util.ParseError(
            'third intro paragraph has the wrong number of lines: '
            f'got {len(lines)}, expected 3')

    heal_skill_id = parse_text.parse_words_seq(lines[0], 'heal skill', 1)[0]
    utility_skill_ids = (
        parse_text.parse_words_seq(lines[1], 'utility skills', 3))
    elite_skill_id = parse_text.parse_words_seq(lines[2], 'elite skill', 1)[0]
    return build.Skills(
        lookup_skill(heal_skill_id, api_storage),
        [lookup_skill(s, api_storage) for s in utility_skill_ids],
        lookup_skill(elite_skill_id, api_storage))


def parse (lines, meta, api_storage):
    paragraphs = list(util.group_paragraphs(
        util.strip_empty_lines(lines, inner='collapse')))
    if len(paragraphs) < 3:
        raise util.ParseError('intro is incomplete')

    setup = parse_setup(paragraphs[-2], meta, api_storage)
    return build.Intro(
        parse_url(paragraphs[0]),
        parse_description(paragraphs[1:-2]),
        setup['gear'],
        setup['traits'],
        parse_skills(paragraphs[-1], meta, api_storage),
        setup['profession options'])



import re


class ParseError (ValueError):
    pass


_title_regex = re.compile(r'(#+)([^#].*)?$')
def _parse_section_title (line):
    match = _title_regex.match(line)

    if match is not None:
        level = len(match[1])
        title = '' if match[2] is None else match[2].strip()
        return (level, title)


# rstrip()s lines
def split_sections (lines, first_title):
    lines_iter = iter(lines)
    next_title = first_title

    def _iter_section ():
        nonlocal next_title
        next_title = None

        for line in lines_iter:
            title = _parse_section_title(line)
            if title is None:
                yield line.rstrip()
            else:
                title_level, title_string = title
                if title_level == 1:
                    next_title = title_string
                    break
                else:
                    yield '{} {}'.format('#' * title_level, title_string)

    while next_title is not None:
        yield (next_title, _iter_section())


def strip_empty_lines (lines, leading=True, trailing=True, inner=None):
    in_leading = True
    empty_count = 0

    for line in lines:
        if not line:
            if not (leading and in_leading):
                empty_count += 1

        else:
            if inner == 'all' and not in_leading:
                pass
            elif inner == 'collapse' and not in_leading:
                if empty_count > 0:
                    yield ''
            else: # inner is None or in_leading
                for i in range(empty_count):
                    yield ''

            empty_count = 0
            in_leading = False
            yield line

    if not trailing:
        for i in range(empty_count):
            yield ''


def group_paragraphs (lines):
    paragraph = []
    for line in lines:
        if line:
            paragraph.append(line)
        else:
            yield paragraph
            paragraph = []
    yield paragraph



import setuptools

with open('README.md', 'r') as f:
    readme_text = f.read()

setuptools.setup(
    name='gw2buildutil',
    version='0.2',
    author='Joseph Lansdowne',
    author_email='ikn@ikn.org.uk',
    description='Python library for working with Guild Wars 2 builds',
    long_description=readme_text,
    long_description_content_type='text/markdown',
    url='http://ikn.org.uk/lib/gw2buildutil',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ]
)



import re

from .. import build
from . import util


word_pattern = r'[\w"\'!\-]+'
words_pattern = f'{word_pattern}' r'( ' f'{word_pattern}' r')*'


def sep_pattern (sep, item_pattern):
    return (f'{item_pattern}'
            r'(' f'{re.escape(sep)}{item_pattern}' r')*')


_gear_groups_pattern = re.compile('^'
    f'{sep_pattern(" + ", words_pattern)}'
    '$')

def parse_gear_groups (text):
    if not text:
        return []
    if _gear_groups_pattern.match(text) is None:
        raise util.ParseError('gear definition doesn\'t match expected '
                              f'format: {repr(text)}')

    return [
        build.GearGroups.from_id(group_text)
        for group_text in text.split(' + ')
    ]

_words_pattern = re.compile(words_pattern)
def parse_words_seq (text, defn_label, num=None):
    if _words_pattern.match(text) is None:
        raise util.ParseError(f'{defn_label} definition doesn\'t match '
                              f'expected format: {repr(text)}')

    items = text.split(', ')
    if num is not None and len(items) != num:
        raise util.ParseError(f'expected {num} values in {defn_label} '
                              f'definition, got {len(items)}')

    return items



import logging
import base64
import struct

from . import build as gw2build, api

# specification: https://en-forum.guildwars2.com/discussion/94395/api-updates-december-18-2019

logger = logging.getLogger(__name__)

_READER_FORMATS = {
    1: 'B',
    2: '<H',
    4: '<I',
}
_SKILLS_SIZE = 2 * 5 * 2
_LEGEND_SKILLS_SIZE = 2 * 3 * 2
_LEGENDS_TOTAL_SIZE = 2 * 2 * 1 + _LEGEND_SKILLS_SIZE
_PETS_SIZE = 2 * 2 * 1
_SUFFIX_SIZE = max(_LEGENDS_TOTAL_SIZE, _PETS_SIZE)


class ParseError (ValueError):
    pass


class _Reader:
    def __init__ (self, data):
        logger.debug(f'template code bytes: {" ".join(map(str, data))}')
        self._data = data
        self._offset = 0
        self._size = len(data)

    def skip (self, size=1):
        logger.debug(f'skip {self._offset}:{size}')
        self._offset += size

    def read (self, size=1):
        if self._offset + size > self._size:
            raise ParseError(f'input too short: {self._size} bytes, '
                             f'want {self._offset + size}')

        fmt = _READER_FORMATS[size]
        value = struct.unpack_from(fmt, self._data, self._offset)[0]
        logger.debug(f'read {self._offset}:{size} -> {value}')
        self._offset += size
        return value


def _parse_profession (reader, api_storage):
    try:
        return api_storage.from_id(api.entity.Profession, reader.read())
    except KeyError:
        raise ParseError('invalid profession ID')


def _parse_spec (reader, api_storage):
    spec_api_id = reader.read()
    if spec_api_id == 0:
        reader.skip()
        return None
    try:
        spec = api_storage.from_api_id(api.entity.Specialisation, spec_api_id)
    except KeyError:
        raise ParseError('invalid specialisation ID')

    traits_data = reader.read()
    choices_indices = [
        traits_data & 0b11,
        (traits_data & 0b1100) >> 2,
        (traits_data & 0b110000) >> 4,
    ]
    choices = [
        None if n == 0 else gw2build.TraitChoices.from_index(n - 1)
        for n in choices_indices]
    return gw2build.SpecialisationChoices(spec, choices)


def _parse_traits (reader, api_storage):
    return gw2build.Traits([_parse_spec(reader, api_storage) for i in range(3)])


def _parse_skill (reader, profession, api_storage):
    build_id = reader.read(2)
    if build_id == 0:
        return None
    else:
        return api.entity.Skill.from_build_id(profession, build_id, api_storage)


def _parse_skills (reader, profession, api_storage):
    skills = [_parse_skill(reader, profession, api_storage) for i in range(10)]
    terrestrial_skills = gw2build.Skills(
        skills[0], [skills[2], skills[4], skills[6]], skills[8])
    aquatic_skills = gw2build.Skills(
        skills[1], [skills[3], skills[5], skills[7]], skills[9])
    return (terrestrial_skills, aquatic_skills)


def _parse_revenant_legend (reader, api_storage):
    build_id = reader.read()
    if build_id == 0:
        return None
    else:
        try:
            return api_storage.from_id(api.entity.RevenantLegend, build_id)
        except KeyError:
            raise ParseError('invalid legend ID')


def _parse_revenant_skills (reader, api_storage):
    legends = [_parse_revenant_legend(reader, api_storage) for i in range(4)]
    reader.skip(_LEGEND_SKILLS_SIZE)
    return (gw2build.RevenantSkills(legends[:2]),
            gw2build.RevenantSkills(legends[2:]))


def _parse_ranger_pet (reader, api_storage):
    api_id = reader.read()
    if api_id == 0:
        return None
    else:
        try:
            return api_storage.from_api_id(api.entity.RangerPet, api_id)
        except KeyError:
            raise ParseError('invalid pet ID')


def _parse_ranger_pets (reader, api_storage):
    pets = [_parse_ranger_pet(reader, api_storage) for i in range(4)]
    return (gw2build.RangerPets(pets[:2]), gw2build.RangerPets(pets[2:]))


def parse (code, api_storage):
    if len(code) < 3 or code[:2] != '[&' or code[-1] != ']':
        raise ParseError('invalid format')
    reader = _Reader(base64.b64decode(code[2:-1]))
    if reader.read() != 0xd:
        raise ParseError('invalid format')

    prof = _parse_profession(reader, api_storage)
    traits = _parse_traits(reader, api_storage)
    if prof.id_ == 'revenant':
        reader.skip(_SKILLS_SIZE)
    else:
        skills, aquatic_skills = _parse_skills(reader, prof, api_storage)

    if prof.id_ == 'revenant':
        skills, aquatic_skills = _parse_revenant_skills(reader, api_storage)
    if prof.id_ == 'ranger':
        pets, aquatic_pets = _parse_ranger_pets(reader, api_storage)
        prof_opts = gw2build.RangerOptions(pets, aquatic_pets)
    else:
        prof_opts = None

    elite_spec = None
    for spec in traits.specs:
        if spec is not None and spec.spec.is_elite:
            elite_spec = spec
    meta = gw2build.BuildMetadata(None, prof, elite_spec)
    intro = gw2build.Intro(None, None, None,
                        traits, skills, prof_opts, aquatic_skills)
    return gw2build.Build(meta, intro)


def _render_profession (build):
    return [build.metadata.profession.build_id]


def _render_spec (spec_choices):
    if spec_choices is None:
        return [0, 0]

    choices_data = [0 if choice is None else (choice.value.index + 1)
                    for choice in spec_choices.choices]
    return [
        spec_choices.spec.api_id,
        choices_data[0] | (choices_data[1] << 2) | (choices_data[2] << 4),
    ]


def _render_traits (build):
    data = []
    # elite spec must come last
    specs = sorted(build.intro.traits.specs,
                   key=lambda spec_choices: spec_choices.spec.is_elite)
    for spec_choices in specs:
        data.extend(_render_spec(spec_choices))
    return data


def _render_skill (skill):
    return struct.pack(_READER_FORMATS[2],
                       0 if skill is None else skill.build_id)


def _render_skills (build):
    aquatic_skills = (gw2build.Skills(None, (None, None, None), None)
                      if build.intro.aquatic_skills is None
                      else build.intro.aquatic_skills)

    skills = [
        build.intro.skills.heal,
        aquatic_skills.heal,
        build.intro.skills.utilities[0],
        aquatic_skills.utilities[0],
        build.intro.skills.utilities[1],
        aquatic_skills.utilities[1],
        build.intro.skills.utilities[2],
        aquatic_skills.utilities[2],
        build.intro.skills.elite,
        aquatic_skills.elite,
    ]

    data = []
    for skill in skills:
        data.extend(_render_skill(skill))
    return data


def _render_revenant_legends (build):
    data = []

    for skills in (build.intro.skills, build.intro.aquatic_skills):
        if skills is None:
            data.extend([0] * 2 * 1)
        else:
            for legend in skills.legends:
                data.append(0 if legend is None else legend.build_id)

    # legend skill order not currently stored in the build
    data.extend([0] * _LEGEND_SKILLS_SIZE)

    return data


def _render_ranger_pets (build):
    data = []
    for pets in (
        build.intro.profession_options.pets,
        build.intro.profession_options.aquatic_pets,
    ):
        if pets is None:
            data.extend([0] * 2)
        else:
            for pet in pets.pets:
                data.append(0 if pet is None else pet.api_id)
    return data


def render (build):
    data = []
    data.append(0xd)
    data.extend(_render_profession(build))
    data.extend(_render_traits(build))
    if build.metadata.profession.id_ == 'revenant':
        data.extend([0] * _SKILLS_SIZE)
    else:
        data.extend(_render_skills(build))

    if build.metadata.profession.id_ == 'revenant':
        data.extend(_render_revenant_legends(build))
        data.extend([0] * (_SUFFIX_SIZE - _LEGENDS_TOTAL_SIZE))
    elif build.metadata.profession.id_ == 'ranger':
        data.extend(_render_ranger_pets(build))
        data.extend([0] * (_SUFFIX_SIZE - _PETS_SIZE))
    else:
        data.extend([0] * _SUFFIX_SIZE)

    code = base64.b64encode(bytes(data)).decode('ascii')
    return f'[&{code}]'



import collections
import enum
import itertools
import math
import sys
import json
import pprint

from .build import Build, BoonTargets


class Configuration:
    def __init__ (
        self, target_buffs, target_uptime, overstack_uptime=None,
        uptime_comparison_tolerance=.01
    ):
        self.target_buffs = target_buffs
        self.target_uptime = target_uptime
        self.overstack_uptime = overstack_uptime
        self.uptime_comparison_tolerance = uptime_comparison_tolerance
        self.max_group_size = 5


class BuffUptime:
    def __init__ (self, config, buff, target, uptime):
        self.buff = buff
        self.target = target
        self._uptime = uptime
        self.uptime = min(config.target_uptime,
                          config.target_uptime if uptime is None else uptime)

    def __str__ (self):
        prefix = '' if self._uptime is None else f'{self.uptime * 100:.0f}% '
        return f'{prefix}{self.target.value.name} {self.buff.value.name}'

    def __repr__ (self):
        return f'BuffUptime<{str(self)}>'

    def uptime_for (self, target):
        if self.target == BoonTargets.PARTY:
            if target == BoonTargets.PARTY:
                return self.uptime
            else:
                return 0
        else: # self.target == BoonTargets.SQUAD
            return self.uptime

    @staticmethod
    def from_boon_uptime (boon_uptime, config):
        uptime = (None if boon_uptime.uptime_percent is None
                  else boon_uptime.uptime_percent / 100)
        return BuffUptime(config, boon_uptime.boon, boon_uptime.target, uptime)

class Role:
    _last_id = 0

    def __init__ (self, buff_uptimes, build_names):
        Role._last_id += 1
        self.id_ = str(Role._last_id)
        self.buff_uptimes = list(buff_uptimes)
        self.build_names = set(build_names)
        self.providing_roles = set()

    def __str__ (self):
        return ', '.join(str(uptime) for uptime in self.buff_uptimes)

    def __repr__ (self):
        return f'Role<{self.id_}: {str(self)}>'

    def _add_providing_role (self, role):
        self.providing_roles.add(role)

    def provides_buff (self, buff):
        return any(uptime.buff == buff for uptime in self.buff_uptimes)

    def uptime (self, buff, target):
        return max((uptime.uptime_for(target)
                    for uptime in self.buff_uptimes
                    if uptime.buff == buff), default=0)

    @staticmethod
    def _simplify_uptimes (uptimes):
        simplified_uptimes = collections.defaultdict(
            lambda: collections.defaultdict(lambda: 0))
        for uptime in uptimes:
            for target in BoonTargets:
                simplified_uptimes[uptime.buff][target] = max(
                    simplified_uptimes[uptime.buff][target],
                    uptime.uptime_for(target))
        return simplified_uptimes

    @staticmethod
    def _uptimes_equal (uptimes1, uptimes2, config):
        uptimes_lookup1 = Role._simplify_uptimes(uptimes1)
        uptimes_lookup2 = Role._simplify_uptimes(uptimes2)
        for buff in set(uptimes_lookup1).union(uptimes_lookup2):
            for target in BoonTargets:
                uptime1 = uptimes_lookup1[buff][target]
                uptime2 = uptimes_lookup2[buff][target]
                if abs(uptime1 - uptime2) > config.uptime_comparison_tolerance:
                    return False
        return True

    @staticmethod
    def _uptimes_greater (uptimes1, uptimes2, config):
        uptimes_lookup1 = Role._simplify_uptimes(uptimes1)
        uptimes_lookup2 = Role._simplify_uptimes(uptimes2)
        for buff in set(uptimes_lookup1).union(uptimes_lookup2):
            for target in BoonTargets:
                uptime1 = uptimes_lookup1[buff][target]
                uptime2 = uptimes_lookup2[buff][target]
                if uptime1 - uptime2 < -config.uptime_comparison_tolerance:
                    return False
        return True

    @staticmethod
    def list_from_builds (builds, config):
        builds_uptimes = {}
        for name, build in builds.items():
            if build.boon_notes is None:
                continue
            for i, variant in enumerate(build.boon_notes.boon_uptime_variants):
                builds_uptimes[(name, i)] = [
                    BuffUptime.from_boon_uptime(buff_uptime, config)
                    for buff_uptime in variant.boon_uptimes]

        eq_builds = collections.defaultdict(lambda: set())
        for (name1, uptimes1), (name2, uptimes2) in \
            itertools.combinations(builds_uptimes.items(), 2) \
        :
            if Role._uptimes_equal(uptimes1, uptimes2, config):
                eq_builds[name1].add(name2)
                eq_builds[name2].add(name1)

        builds_groups = {}
        for name in builds_uptimes:
            build_group = eq_builds[name]
            build_group.add(name)
            for eq_name in build_group:
                if eq_name in builds_groups:
                    builds_groups[eq_name].update(build_group)
                    break
            else:
                builds_groups[name] = build_group

        roles = [
            Role(builds_uptimes[next(iter(build_group))],
                 [name[0] for name in build_group])
            for build_group in builds_groups.values()]

        providing_roles = collections.defaultdict(lambda: set())
        for role1, role2 in itertools.permutations(roles, 2):
            if (role1 is not role2 and
                Role._uptimes_greater(
                    role1.buff_uptimes, role2.buff_uptimes, config)
            ):
                role2._add_providing_role(role1)

        return roles


class _SimpleComposition:
    def __init__ (self, config, group1, group2):
        self._config = config
        self.group1 = list(group1)
        self.group1_counter = collections.Counter(self.group1)
        self.group2 = list(group2)
        self.group2_counter = collections.Counter(self.group2)
        self._roles_counter = collections.Counter(self.roles)

    def __str__ (self):
        return ' | '.join(' '.join(role.id_ for role in group)
                         for group in (self.group1, self.group2))

    def __repr__ (self):
        return f'_SimpleComposition<{str(self)}>'

    @property
    def roles (self):
        return self.group1 + self.group2

    def __eq__ (self, other):
        return self._roles_counter == other._roles_counter

    def __len__ (self):
        return len(self.group1) + len(self.group2)

    def provides (self, other):
        remaining_roles = self._roles_counter - other._roles_counter
        remaining_other_roles = other._roles_counter - self._roles_counter

        # match roles with only 1 providing role
        for other_role in remaining_other_roles:
            our_providing_roles = [role for role in remaining_roles
                                   if role in other_role.providing_roles]
            if len(our_providing_roles) == 1:
                # `counter[i] -=` can produce negative counts
                reduction = min(remaining_other_roles[other_role],
                                remaining_roles[our_providing_roles[0]])
                remaining_roles[our_providing_roles[0]] -= reduction
                remaining_other_roles[other_role] -= reduction

        remaining_roles_seq = list(remaining_roles.elements())
        remaining_other_roles_seq = list(remaining_other_roles.elements())
        if len(remaining_roles_seq) < len(remaining_other_roles_seq):
            return False

        # check all possible reorderings to match the rest
        for roles_ordering in itertools.permutations(
            remaining_roles_seq, len(remaining_other_roles_seq)
        ):
            if (all(role in other_role.providing_roles
                    for role, other_role
                    in zip(roles_ordering, remaining_other_roles_seq))
            ):
                return True
        return False

    def uptime (self, buff):
        result = []
        for group in (1, 2):
            own_group = self.group1 if group == 1 else self.group2
            own_group_uptime = sum(
                role.uptime(buff, BoonTargets.PARTY)
                for role in own_group)
            off_group = self.group2 if group == 1 else self.group1
            off_group_uptime = sum(
                role.uptime(buff, BoonTargets.SQUAD)
                for role in off_group)
            result.append(own_group_uptime + off_group_uptime)
        return result

    def overstack (self):
        if self._config.overstack_uptime is None:
            return False

        for buff in self._config.target_buffs:
            for uptime in self.uptime(buff):
                if uptime >= self._config.overstack_uptime:
                    return True

        return False


    @staticmethod
    def empty (config):
        return _SimpleComposition(config, [], [])


class Composition:
    def __init__ (self, comps):
        self.compositions = list(comps)

    @property
    def roles (self):
        return self.compositions[0].roles

    def __str__ (self):
        return ' '.join('[' + str(comp) + ']' for comp in self.compositions)

    def __repr__ (self):
        return f'Composition<{str(self)}>'


def _simplify_compositions (config, comps):
    comps = list(comps)

    # group together comps that use the same roles
    grouped_comps = []
    i = 0
    while i < len(comps):
        comp1 = comps[i]
        groupings = [comp1]
        grouped_comps.append(groupings)
        j = i + 1
        while j < len(comps):
            comp2 = comps[j]
            if comp2 == comp1:
                groupings.append(comp2)
                comps.pop(j) # now we can skip over comp2 in the outer loop
            else:
                j += 1
        i += 1

    # filter out comps which are the same as another comp, with roles replaced
    # with other roles with strictly better uptimes (eg. don't use an
    # inspiration chrono when you can use a firebrand)

    # sort so we process smaller comps first, which can be compared faster
    grouped_comps.sort(key=lambda groupings: len(groupings[0]))
    simplified_grouped_comps = []
    for new_groupings in grouped_comps:
        if new_groupings[0].overstack():
            continue

        j = 0
        while j < len(simplified_grouped_comps):
            existing_groupings = simplified_grouped_comps[j]
            if new_groupings[0].provides(existing_groupings[0]):
                break
            elif existing_groupings[0].provides(new_groupings[0]):
                simplified_grouped_comps.pop(j)
            else:
                j += 1
        else:
            simplified_grouped_comps.append(new_groupings)

    for groupings in simplified_grouped_comps:
        normalised_groupings = []
        for comp in groupings:
            for existing_comp in normalised_groupings:
                is_same = (
                    comp.group1_counter == existing_comp.group1_counter and
                        comp.group2_counter == existing_comp.group2_counter)
                is_mirrored = (
                    comp.group1_counter == existing_comp.group2_counter and
                    comp.group2_counter == existing_comp.group1_counter)
                if is_same or is_mirrored:
                    break
            else:
                normalised_groupings.append(comp)
        yield Composition(normalised_groupings)


# base_comp must contain exactly 1 grouping
# target_uptime has tolerance applied
def _generate_compositions_for_buff (
    base_comp, roles_group1, roles_group2, buff, config
):
    uptime_group1, uptime_group2 = base_comp.uptime(buff)
    needed_group1 = config.target_uptime - uptime_group1
    allowed_group1 = (1000 if config.overstack_uptime is None
                      else config.overstack_uptime - uptime_group1)
    needed_group2 = config.target_uptime - uptime_group2
    allowed_group2 = (1000 if config.overstack_uptime is None
                      else config.overstack_uptime - uptime_group2)
    if ((needed_group1 < config.uptime_comparison_tolerance and
         needed_group2 < config.uptime_comparison_tolerance)
    ):
        yield base_comp
        return

    # when recurring, we don't pass through roles we've already checked, because
    # that will just produce reorderings of the same compositions
    for role_index, role in enumerate(roles_group1):
        provides_group = role.uptime(buff, BoonTargets.PARTY)
        provides_off_group = role.uptime(buff, BoonTargets.SQUAD)
        max_usable_group = (0 if provides_group == 0
                            else math.ceil(needed_group1 / provides_group))
        max_usable_off_group = (
            0 if provides_off_group == 0
            else math.ceil(needed_group2 / provides_off_group))
        max_allowed = min(
            config.max_group_size - len(base_comp.group1),
            # doesn't catch cases where we overstack on a different buff, but
            # all comps are checked again for overstack at the end
            (config.max_group_size if provides_group == 0
             else int(allowed_group1 / provides_group)),
            (config.max_group_size if provides_off_group == 0
             else int(allowed_group2 / provides_off_group)))
        max_to_use = min(max_allowed,
                         max(max_usable_group, max_usable_off_group))
        for num_used in range(1, max_to_use + 1):
            comp = _SimpleComposition(config,
                                      base_comp.group1 + [role] * num_used,
                                      base_comp.group2)
            yield from _generate_compositions_for_buff(
                comp, roles_group1[role_index + 1:], roles_group2, buff, config)

    # if groups are identical, adding to group 2 just produces mirrored comps
    if base_comp.group1_counter == base_comp.group2_counter:
        return

    for role_index, role in enumerate(roles_group2):
        provides_group = role.uptime(buff, BoonTargets.PARTY)
        provides_off_group = role.uptime(buff, BoonTargets.SQUAD)
        max_usable_group = (0 if provides_group == 0
                            else math.ceil(needed_group2 / provides_group))
        max_usable_off_group = (
            0 if provides_off_group == 0
            else math.ceil(needed_group1 / provides_off_group))
        max_allowed = min(
            config.max_group_size - len(base_comp.group2),
            # doesn't catch cases where we overstack on a different buff, but
            # all comps are checked again for overstack at the end
            (config.max_group_size if provides_group == 0
             else int(allowed_group2 / provides_group)),
            (config.max_group_size if provides_off_group == 0
             else int(allowed_group1 / provides_off_group)))
        max_to_use = min(max_allowed,
                         max(max_usable_group, max_usable_off_group))
        for num_used in range(1, max_to_use + 1):
            comp = _SimpleComposition(config,
                                      base_comp.group1,
                                      base_comp.group2 + [role] * num_used)
            yield from _generate_compositions_for_buff(
                comp, roles_group1, roles_group2[role_index + 1:], buff, config)


def generate_compositions (roles, config):
    roles_by_buff = {buff: [role for role in roles if role.provides_buff(buff)]
                     for buff in config.target_buffs}
    def generate (target_buffs):
        if len(target_buffs) > 1:
            base_comps = generate(target_buffs[:-1])
        else:
            base_comps = [_SimpleComposition.empty(config)]
        for base_comp in base_comps:
            roles_for_buff = roles_by_buff[target_buffs[-1]]
            yield from _generate_compositions_for_buff(
                base_comp,
                roles_for_buff,
                roles_for_buff,
                target_buffs[-1],
                config)

    ordered_target_buffs = sorted(config.target_buffs,
                                  key=lambda buff: len(roles_by_buff[buff]))
    return _simplify_compositions(
        config, list(generate(list(ordered_target_buffs))))
